@@ -27,6 +27,7 @@
 #include <syslog.h>
 #include <sys/types.h>
 #include <sys/wait.h>
+#include <sys/stat.h>
 #include <fcntl.h>
 #include "gdm.h"
 #include "misc.h"
@@ -40,6 +41,19 @@ extern char **stored_argv;
 extern int stored_argc;
 extern char *stored_path;
 
+static int screenx = 0;
+static int screeny = 0;
+static int screenwidth = 0;
+static int screenheight = 0;
+
+static void
+setup_cursor (GdkCursorType type)
+{
+	GdkCursor *cursor = gdk_cursor_new (type);
+	gdk_window_set_cursor (gdk_get_default_root_window (), cursor);
+	gdk_cursor_unref (cursor);
+}
+
 static gboolean
 gdm_event (GSignalInvocationHint *ihint,
 	   guint		n_param_values,
@@ -65,32 +79,91 @@ gdm_event (GSignalInvocationHint *ihint,
 		event->button.button = 1;
 
 	return TRUE;
-}      
+}
+
+static void
+get_screen_size (GdmDisplay *d)
+{
+	if (d != NULL) {
+		screenx = d->screenx;
+		screeny = d->screeny;
+		screenwidth = d->screenwidth;
+		screenheight = d->screenheight;
+	}
+
+	if (screenwidth <= 0)
+		screenwidth = gdk_screen_width ();
+	if (screenheight <= 0)
+		screenheight = gdk_screen_height ();
+}
+
+static void
+center_window (GtkWidget *window)
+{
+	GtkRequisition req;
+
+	gtk_widget_size_request (window, &req);
+
+	gtk_window_move (GTK_WINDOW (window),
+			 screenx +
+			 (screenwidth / 2) -
+			 (req.width / 2),
+			 screeny +
+			 (screenheight / 2) -
+			 (req.height / 2));
+}
 
 static void
 show_errors (GtkWidget *button, gpointer data)
 {
 	const char *file = data;
+	char *details;
 	FILE *fp;
+	struct stat s;
 	GtkWidget *sw;
 	GtkWidget *label;
+	GtkWidget *dlg = gtk_widget_get_toplevel (button);
 	GtkWidget *parent = button->parent;
 	GString *gs = g_string_new (NULL);
+	gboolean valid_utf8 = TRUE;
+
+	fp = NULL;
+	if (stat (file, &s) == 0) {
+		if (S_ISREG (s.st_mode))
+			fp = fopen (file, "r");
+		else {
+			g_string_printf (gs, "%s not a regular file!\n", file);
+		}
+	}
 
-	fp = fopen (file, "r");
 	if (fp != NULL) {
-		char buf[256];
-		while (fgets (buf, sizeof (buf), fp))
+		char buf[128];
+		int lines = 0;
+		while (fgets (buf, sizeof (buf), fp)) {
+			if ( ! g_utf8_validate (buf, -1, NULL))
+				valid_utf8 = FALSE;
 			g_string_append (gs, buf);
+			/* cap lines at 100 */
+			if (lines ++ > 100) {
+				g_string_append (gs, "\n... File too long to display ...\n");
+				break;
+			}
+		}
 		fclose (fp);
 	} else {
-		g_string_printf (gs, _("%s could not be opened"), file);
+		char  *loc;
+		loc = gdm_locale_to_utf8 (_("%s could not be opened"));
+		g_string_append_printf (gs, loc, file);
+		g_free (loc);
 	}
 
 	gtk_widget_destroy (button);
 
 	sw = gtk_scrolled_window_new (NULL, NULL);
-	gtk_widget_set_size_request (sw, 200, 150);
+	if (gdk_screen_width () >= 800)
+		gtk_widget_set_size_request (sw, 500, 150);
+	else
+		gtk_widget_set_size_request (sw, 200, 150);
 	gtk_widget_show (sw);
 
 	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
@@ -100,14 +173,22 @@ show_errors (GtkWidget *button, gpointer data)
 	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),
 					     GTK_SHADOW_IN);
 
-	label = gtk_label_new (gs->str);
+	details = g_string_free (gs, FALSE);
+
+	if ( ! valid_utf8) {
+		char *tmp = gdm_locale_to_utf8 (details);
+		g_free (details);
+		details = tmp;
+	}
+
+	label = gtk_label_new (details);
 	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (sw), label);
 	gtk_widget_show (label);
 
 	gtk_box_pack_start (GTK_BOX (parent),
 			    sw, TRUE, TRUE, 0);
 
-	g_string_free (gs, TRUE);
+	center_window (dlg);
 }
 
 void
@@ -120,31 +201,40 @@ gdm_error_box_full (GdmDisplay *d, GtkMessageType type, const char *error,
 
 	if (pid == 0) {
 		guint sid;
-		int i;
 		int argc = 1;
 		char **argv;
 		GtkWidget *dlg;
 		GtkWidget *button;
-		GtkRequisition req;
-		int screenx = 0;
-		int screeny = 0;
-		int screenwidth = 0;
-		int screenheight = 0;
 		char *loc;
+		char *display;
+		char *xauthority;
 
 		closelog ();
 
-		for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-			close(i);
+		gdm_close_all_descriptors (0 /* from */, -1 /* except */);
 
 		/* No error checking here - if it's messed the best response
 		 * is to ignore & try to continue */
-		open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-		open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
 
 		gdm_desetuid ();
 
+		display = g_strdup (g_getenv ("DISPLAY"));
+		xauthority = g_strdup (g_getenv ("XAUTHORITY"));
+
+		/* restore initial environment */
+		gdm_restoreenv ();
+
+		if (display != NULL)
+			gnome_setenv ("DISPLAY", display, TRUE);
+		if (xauthority != NULL)
+			gnome_setenv ("XAUTHORITY", xauthority, TRUE);
+		/* sanity env stuff */
+		gnome_setenv ("SHELL", "/bin/bash", TRUE);
+		gnome_setenv ("HOME", "/", TRUE);
+
 		openlog ("gdm", LOG_PID, LOG_DAEMON);
 
 		argv = g_new0 (char *, 2);
@@ -152,19 +242,9 @@ gdm_error_box_full (GdmDisplay *d, GtkMessageType type, const char *error,
 
 		gtk_init (&argc, &argv);
 
-		if (d != NULL) {
-			screenx = d->screenx;
-			screeny = d->screeny;
-			screenwidth = d->screenwidth;
-			screenheight = d->screenheight;
-		}
+		get_screen_size (d);
 
-		if (screenwidth <= 0)
-			screenwidth = gdk_screen_width ();
-		if (screenheight <= 0)
-			screenheight = gdk_screen_height ();
-
-		loc = g_locale_to_utf8 (error, -1, NULL, NULL, NULL);
+		loc = gdm_locale_to_utf8 (error);
 
 		dlg = gtk_message_dialog_new (NULL /* parent */,
 					      0 /* flags */,
@@ -173,10 +253,10 @@ gdm_error_box_full (GdmDisplay *d, GtkMessageType type, const char *error,
 					      "%s",
 					      loc);
 		g_free (loc);
+		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 
 		if (details_label != NULL) {
-			loc = g_locale_to_utf8 (details_label,
-						-1, NULL, NULL, NULL);
+			loc = gdm_locale_to_utf8 (details_label);
 			button = gtk_button_new_with_label (loc);
 			g_free (loc);
 
@@ -202,15 +282,7 @@ gdm_error_box_full (GdmDisplay *d, GtkMessageType type, const char *error,
 					    NULL /* data */,
 					    NULL /* destroy_notify */);
 
-		gtk_widget_size_request (dlg, &req);
-
-		gtk_window_move (GTK_WINDOW (dlg),
-				 screenx +
-				 (screenwidth / 2) -
-				 (req.width / 2),
-				 screeny +
-				 (screenheight / 2) -
-				 (req.height / 2));
+		center_window (dlg);
 
 		gtk_widget_grab_focus (button);
 
@@ -226,6 +298,8 @@ gdm_error_box_full (GdmDisplay *d, GtkMessageType type, const char *error,
 			gdk_error_trap_pop ();
 		}
 
+		setup_cursor (GDK_LEFT_PTR);
+
 		gtk_dialog_run (GTK_DIALOG (dlg));
 
 		XSetInputFocus (GDK_DISPLAY (),
@@ -268,30 +342,39 @@ gdm_failsafe_question (GdmDisplay *d,
 	pid = gdm_fork_extra ();
 	if (pid == 0) {
 		guint sid;
-		int i;
 		int argc = 1;
 		char **argv;
 		GtkWidget *dlg, *label, *entry;
-		GtkRequisition req;
-		int screenx = 0;
-		int screeny = 0;
-		int screenwidth = 0;
-		int screenheight = 0;
 		char *loc;
+		char *display;
+		char *xauthority;
 
 		closelog ();
 
-		for (i = 0; i < sysconf (_SC_OPEN_MAX); i++)
-			close(i);
+		gdm_close_all_descriptors (0 /* from */, p[1] /* except */);
 
 		/* No error checking here - if it's messed the best response
 		 * is to ignore & try to continue */
-		open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-		open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
 
 		gdm_desetuid ();
 
+		display = g_strdup (g_getenv ("DISPLAY"));
+		xauthority = g_strdup (g_getenv ("XAUTHORITY"));
+
+		/* restore initial environment */
+		gdm_restoreenv ();
+
+		if (display != NULL)
+			gnome_setenv ("DISPLAY", display, TRUE);
+		if (xauthority != NULL)
+			gnome_setenv ("XAUTHORITY", xauthority, TRUE);
+		/* sanity env stuff */
+		gnome_setenv ("SHELL", "/bin/bash", TRUE);
+		gnome_setenv ("HOME", "/", TRUE);
+
 		openlog ("gdm", LOG_PID, LOG_DAEMON);
 
 		argv = g_new0 (char *, 2);
@@ -299,19 +382,9 @@ gdm_failsafe_question (GdmDisplay *d,
 
 		gtk_init (&argc, &argv);
 
-		if (d != NULL) {
-			screenx = d->screenx;
-			screeny = d->screeny;
-			screenwidth = d->screenwidth;
-			screenheight = d->screenheight;
-		}
+		get_screen_size (d);
 
-		if (screenwidth <= 0)
-			screenwidth = gdk_screen_width ();
-		if (screenheight <= 0)
-			screenheight = gdk_screen_height ();
-
-		loc = g_locale_to_utf8 (question, -1, NULL, NULL, NULL);
+		loc = gdm_locale_to_utf8 (question);
 
 		dlg = gtk_dialog_new_with_buttons (loc,
 						   NULL /* parent */,
@@ -323,9 +396,11 @@ gdm_failsafe_question (GdmDisplay *d,
 				  G_CALLBACK (gtk_true), NULL);
 
 		label = gtk_label_new (loc);
+		gtk_widget_show_all (label);
 		gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox),
 				    label, FALSE, FALSE, 0);
 		entry = gtk_entry_new ();
+		gtk_widget_show_all (entry);
 		gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox),
 				    entry, FALSE, FALSE, 0);
 		if ( ! echo)
@@ -342,15 +417,7 @@ gdm_failsafe_question (GdmDisplay *d,
 					    NULL /* data */,
 					    NULL /* destroy_notify */);
 
-		gtk_widget_size_request (dlg, &req);
-
-		gtk_window_move (GTK_WINDOW (dlg),
-				 screenx +
-				 (screenwidth / 2) -
-				 (req.width / 2),
-				 screeny +
-				 (screenheight / 2) -
-				 (req.height / 2));
+		center_window (dlg);
 
 		gtk_widget_show_now (dlg);
 
@@ -366,13 +433,18 @@ gdm_failsafe_question (GdmDisplay *d,
 
 		gtk_widget_grab_focus (entry);
 
+		setup_cursor (GDK_LEFT_PTR);
+
 		gtk_dialog_run (GTK_DIALOG (dlg));
 
-		loc = g_locale_from_utf8 (ve_sure_string (gtk_entry_get_text (GTK_ENTRY (entry))),
-					  -1, NULL, NULL, NULL);
+		loc = gdm_locale_from_utf8 (ve_sure_string (gtk_entry_get_text (GTK_ENTRY (entry))));
 
-		printf ("%s", ve_sure_string (loc));
-		fflush (stdout);
+		gdm_fdprintf (p[1], "%s", ve_sure_string (loc));
+
+		XSetInputFocus (GDK_DISPLAY (),
+				PointerRoot,
+				RevertToPointerRoot,
+				CurrentTime);
 
 		_exit (0);
 	} else if (pid > 0) {
@@ -409,32 +481,39 @@ gdm_failsafe_yesno (GdmDisplay *d,
 	pid = gdm_fork_extra ();
 	if (pid == 0) {
 		guint sid;
-		int i;
 		int argc = 1;
 		char **argv;
 		GtkWidget *dlg;
-		GtkRequisition req;
-		int screenx = 0;
-		int screeny = 0;
-		int screenwidth = 0;
-		int screenheight = 0;
 		char *loc;
+		char *display;
+		char *xauthority;
 
 		closelog ();
 
-		for (i = 0; i < sysconf (_SC_OPEN_MAX); i++) {
-			if (p[1] != i)
-				close(i);
-		}
+		gdm_close_all_descriptors (0 /* from */, p[1] /* except */);
 
 		/* No error checking here - if it's messed the best response
 		 * is to ignore & try to continue */
-		open ("/dev/null", O_RDONLY); /* open stdin - fd 0 */
-		open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
-		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
+		gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
+		gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
+		gdm_open_dev_null (O_RDWR); /* open stderr - fd 2 */
 
 		gdm_desetuid ();
 
+		display = g_strdup (g_getenv ("DISPLAY"));
+		xauthority = g_strdup (g_getenv ("XAUTHORITY"));
+
+		/* restore initial environment */
+		gdm_restoreenv ();
+
+		if (display != NULL)
+			gnome_setenv ("DISPLAY", display, TRUE);
+		if (xauthority != NULL)
+			gnome_setenv ("XAUTHORITY", xauthority, TRUE);
+		/* sanity env stuff */
+		gnome_setenv ("SHELL", "/bin/bash", TRUE);
+		gnome_setenv ("HOME", "/", TRUE);
+
 		openlog ("gdm", LOG_PID, LOG_DAEMON);
 
 		argv = g_new0 (char *, 2);
@@ -442,19 +521,9 @@ gdm_failsafe_yesno (GdmDisplay *d,
 
 		gtk_init (&argc, &argv);
 
-		if (d != NULL) {
-			screenx = d->screenx;
-			screeny = d->screeny;
-			screenwidth = d->screenwidth;
-			screenheight = d->screenheight;
-		}
+		get_screen_size (d);
 
-		if (screenwidth <= 0)
-			screenwidth = gdk_screen_width ();
-		if (screenheight <= 0)
-			screenheight = gdk_screen_height ();
-
-		loc = g_locale_to_utf8 (question, -1, NULL, NULL, NULL);
+		loc = gdm_locale_to_utf8 (question);
 
 		dlg = gtk_message_dialog_new (NULL /* parent */,
 					      0 /* flags */,
@@ -462,6 +531,7 @@ gdm_failsafe_yesno (GdmDisplay *d,
 					      GTK_BUTTONS_YES_NO,
 					      "%s",
 					      loc);
+		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 
 		sid = g_signal_lookup ("event",
 				       GTK_TYPE_WIDGET);
@@ -471,15 +541,7 @@ gdm_failsafe_yesno (GdmDisplay *d,
 					    NULL /* data */,
 					    NULL /* destroy_notify */);
 
-		gtk_widget_size_request (dlg, &req);
-
-		gtk_window_move (GTK_WINDOW (dlg),
-				 screenx +
-				 (screenwidth / 2) -
-				 (req.width / 2),
-				 screeny +
-				 (screenheight / 2) -
-				 (req.height / 2));
+		center_window (dlg);
 
 		gtk_widget_show_now (dlg);
 
@@ -493,11 +555,17 @@ gdm_failsafe_yesno (GdmDisplay *d,
 			gdk_error_trap_pop ();
 		}
 
+		setup_cursor (GDK_LEFT_PTR);
+
 		if (gtk_dialog_run (GTK_DIALOG (dlg)) == GTK_RESPONSE_YES)
-			printf ("yes\n");
+			gdm_fdprintf (STDOUT_FILENO, "yes\n");
 		else
-			printf ("no\n");
-		fflush (stdout);
+			gdm_fdprintf (STDOUT_FILENO, "no\n");
+
+		XSetInputFocus (GDK_DISPLAY (),
+				PointerRoot,
+				RevertToPointerRoot,
+				CurrentTime);
 
 		_exit (0);
 	} else if (pid > 0) {
