@@ -64,6 +64,27 @@ setup_cursor (GdkCursorType type)
 	gdk_cursor_unref (cursor);
 }
 
+static void
+setup_window_cursor (GdkCursorType type)
+{
+	GdkCursor *cursor = gdk_cursor_new (type);
+	GtkWidget *setup_dialog = glade_helper_get
+		(xml, "setup_dialog", GTK_TYPE_WINDOW);
+	if (setup_dialog->window)
+		gdk_window_set_cursor (setup_dialog->window, cursor);
+	gdk_cursor_unref (cursor);
+}
+
+static void
+unsetup_window_cursor (void)
+{
+	GtkWidget *setup_dialog = glade_helper_get
+		(xml, "setup_dialog", GTK_TYPE_WINDOW);
+	if (setup_dialog->window)
+		gdk_window_set_cursor (setup_dialog->window, NULL);
+}
+
+
 static void
 update_greeters (void)
 {
@@ -126,6 +147,7 @@ check_update_error:
 						  "login screens.  Not all "
 						  "updates may have taken "
 						  "effect."));
+		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 		gtk_dialog_run (GTK_DIALOG (dlg));
 		gtk_widget_destroy (dlg);
 		shown_error = TRUE;
@@ -393,6 +415,7 @@ setup_user_combo (const char *name, const char *key)
 	GList *users = NULL;
 	struct passwd *pwent;
 	char *str;
+	int cnt;
 
 	str = ve_config_get_string (ve_config_get (GDM_CONFIG_FILE), key);
 
@@ -405,12 +428,20 @@ setup_user_combo (const char *name, const char *key)
 	setpwent ();
 
 	pwent = getpwent();
+	cnt = 0;
 	
 	while (pwent != NULL) {
 		if (pwent->pw_uid >= GdmMinimalUID &&
 		    strcmp (ve_sure_string (str), pwent->pw_name) != 0) {
+			cnt ++;
 			users = g_list_append (users,
 					       g_strdup (pwent->pw_name));
+			/* FIXME: fix properly, see bug #111830 */
+			if (cnt >= 50) {
+				users = g_list_append (users,
+						       g_strdup ("..."));
+				break;
+			}
 		}
 	
 		pwent = getpwent();
@@ -1465,6 +1496,7 @@ install_ok (GtkWidget *button, gpointer data)
 						GTK_MESSAGE_ERROR,
 						GTK_BUTTONS_OK,
 						_("No file selected"));
+		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 		gtk_dialog_run (GTK_DIALOG (dlg));
 		gtk_widget_destroy (dlg);
 		g_free (cwd);
@@ -1493,6 +1525,7 @@ install_ok (GtkWidget *button, gpointer data)
 						_("Not a theme archive\n"
 						  "Details: %s"),
 						error);
+		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 		gtk_dialog_run (GTK_DIALOG (dlg));
 		gtk_widget_destroy (dlg);
 		g_free (filename);
@@ -1515,6 +1548,7 @@ install_ok (GtkWidget *button, gpointer data)
 			 _("Theme directory '%s' seems to be already "
 			   "installed, install again anyway?"),
 			 fname);
+		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 		g_free (fname);
 		if (gtk_dialog_run (GTK_DIALOG (dlg)) != GTK_RESPONSE_YES) {
 			gtk_widget_destroy (dlg);
@@ -1559,6 +1593,7 @@ install_ok (GtkWidget *button, gpointer data)
 						GTK_BUTTONS_OK,
 						_("Some error occured when "
 						  "installing the theme"));
+		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 		gtk_dialog_run (GTK_DIALOG (dlg));
 		gtk_widget_destroy (dlg);
 	}
@@ -1750,6 +1785,7 @@ dialog_response (GtkWidget *dlg, int response, gpointer data)
 			   "are listed here.  You may want to edit %s "
 			   "if you cannot find what you are looking for."),
 			 GDM_CONFIG_FILE);
+		gtk_dialog_set_has_separator (GTK_DIALOG (dlg), FALSE);
 		g_signal_connect (G_OBJECT (dlg), "destroy",
 				  G_CALLBACK (gtk_widget_destroyed),
 				  &dlg);
@@ -1873,6 +1909,88 @@ setup_gui (void)
 					 GDM_KEY_WELCOME);
 }
 
+static gboolean
+get_sensitivity (void)
+{
+	static Atom atom = 0;
+	Display *disp = gdk_x11_get_default_xdisplay ();
+	Window root = gdk_x11_get_default_root_xwindow ();
+	gulong *data;
+	gulong nitems_return;
+	gulong bytes_after_return;
+	Atom type_returned;
+	int format_returned;
+
+	if (atom == 0)
+		atom = XInternAtom (disp, "_GDM_SETUP_INSENSITIVE", False);
+
+	if (XGetWindowProperty (disp,
+				root,
+				atom,
+				0, 1,
+				False,
+				XA_CARDINAL,
+				&type_returned, &format_returned,
+				&nitems_return,
+				&bytes_after_return,
+				(unsigned char **)&data) != Success)
+		return TRUE;
+
+	if (format_returned != 32 ||
+	    data[0] == 0) {
+		XFree (data);
+		return TRUE;
+	} else {
+		XFree (data);
+		return FALSE;
+	}
+}
+
+static void
+update_sensitivity (void)
+{
+	gboolean sensitive = get_sensitivity ();
+	GtkWidget *setup_dialog = glade_helper_get (xml, "setup_dialog",
+						    GTK_TYPE_WINDOW);
+	gtk_widget_set_sensitive (setup_dialog, sensitive);
+	if (sensitive)
+		unsetup_window_cursor ();
+	else
+		setup_window_cursor (GDK_WATCH);
+}
+
+static GdkFilterReturn
+root_window_filter (GdkXEvent *gdk_xevent,
+		    GdkEvent *event,
+		    gpointer data)
+{
+	XEvent *xevent = (XEvent *)gdk_xevent;
+
+	if (xevent->type == PropertyNotify)
+		update_sensitivity ();
+
+	return GDK_FILTER_CONTINUE;
+}
+
+static void
+setup_disable_handler (void)
+{
+	XWindowAttributes attribs = { 0, };
+	Display *disp = gdk_x11_get_default_xdisplay ();
+	Window root = gdk_x11_get_default_root_xwindow ();
+
+	update_sensitivity ();
+
+	/* set event mask for events on root window */
+	XGetWindowAttributes (disp, root, &attribs);
+	XSelectInput (disp, root,
+		      attribs.your_event_mask |
+		      PropertyChangeMask);
+
+	gdk_window_add_filter (gdk_get_default_root_window (),
+			       root_window_filter, NULL);
+}
+
 static gboolean
 gdm_event (GSignalInvocationHint *ihint,
 	   guint		n_param_values,
@@ -1965,6 +2083,7 @@ main (int argc, char *argv[])
 						GTK_MESSAGE_ERROR,
 						GTK_BUTTONS_OK,
 						_("You must be the superuser (root) to configure GDM.\n"));
+		gtk_dialog_set_has_separator (GTK_DIALOG (fatal_error), FALSE);
 		if (RUNNING_UNDER_GDM)
 			setup_cursor (GDK_LEFT_PTR);
 		gtk_dialog_run (GTK_DIALOG (fatal_error));
@@ -1978,8 +2097,11 @@ main (int argc, char *argv[])
 
 	setup_gui ();
 
-	if (RUNNING_UNDER_GDM)
+	if (RUNNING_UNDER_GDM) {
+		setup_disable_handler ();
+
 		setup_cursor (GDK_LEFT_PTR);
+	}
 
 	gtk_main ();
 
